<?php

// Psy Guest Book (psygb)
// Hadrien Nilsson - psydk.org 2005
// 2005-07-03 : v1.1.2
// - Lecture et conversions vers timestamp de dates Ok
// 2005-07-03 : v1.1
// - Écriture atomique d'une unité
// - Ajout du nombre de messages
// 2005-06-20 : v1.0
// - Première release

// Ce code est libre de droit, vous pouvez en faire ce que vous voulez, le copier, le modifier etc.
// Si vous me citez j'en serais ravi mais vous n'êtes pas obligé.
// Attention, malgré tous les soins apportés à sa réalisation, ce bout de code est fourni en l'état,
// sans aucune garantie. En aucun cas l'auteur ne pourra être tenu responsable pour d'éventuels dommages
// liés à l'utilisation de ce code.

// Note : ce fichier est un fichier UTF-8
// Veillez à retirer le BYTE ORDER MARK (EF BB BF) en début de fichier si votre éditeur
// de texte l'ajoute lors d'une édition

///////////////////////////////////////////////////
$k_strDbFileName = 'psygb.db';
$k_strUnitBegin = ">>>" . chr(10);
$k_strUnitEnd = "<<<" . chr(10);
$k_nUnitSize = 512;
$k_nAuthorMaxByteCount = 31;
$k_nMaxPerPage = 10;
$k_nMaxPagePerBlock = 10;
///////////////////////////////////////////////////

function XhtmlSpecialChars($str)
{
	$aForbidden = array('&', '<', '>');
	$aSafe = array('&amp;', '&lt;', '&gt;');
 
	$str = str_replace($aForbidden, $aSafe, $str);
	return $str;
}

function EchoStringHex($str)
{
	$nLength = strlen($str);
	for($i = 0; $i < $nLength; ++$i)
	{
		$nByte = ord($str[$i]);
		printf("%02x ", $nByte);
	}
}

function PadUtf8($str, $nMax)
{
	$nLength = strlen($str);
	if( $nLength <= $nMax)
	{
		// On padde
		return str_pad($str, $nMax);
	}

	// Là ça se complique, faut tronquer selon un code UTF-8 valide

	// Regarde le premier octet qui dépasse
	$nByte = ord($str[$nMax]);
	
	if( ($nByte & 0xC0) == 0xC0)
	{
		// Premier octet d'une séquence, on peut tronquer librement
		return substr($str, 0, $nMax);
	}

	// On est en plein dans une séquence, il va falloir la soustraire
	for($i = $nMax - 1; $i >= 0; --$i)
	{
		// Recherche le premier octet de la séquence
		$nByte = ord($str[$i]);

		if( ($nByte & 0xC0) == 0xC0)
		{
			// Trouvé
			$str2 = substr($str, 0, $i);
			return str_pad($str2, $nMax);
		}
	}

	// UTF-8 invalide
	return '';
}

$strUnitAuthor = '';
$strUnitMessage = '';
$strError = '';

function GetMessageMaxByteCount()
{
	global $k_strUnitBegin, $k_strUnitEnd, $k_nAuthorMaxByteCount, $k_nUnitSize;

	$nSize = strlen($k_strUnitBegin) + strlen($k_strUnitEnd);
	$nSize += 16; // Date\n
	$nSize += 16; // IP\n
	$nSize += $k_nAuthorMaxByteCount + 1; // Author\n
	$nSize += 1; // \n du message

	$nRemains = $k_nUnitSize - $nSize;
	if( $nRemains <= 0)
	{
		echo 'Mauvaise configuration, la taille de l\'unité est trop petite.';
	}
	return $nRemains;
}

function IsoToTimestamp($strIsoDate)
{
	$strYear = substr($strIsoDate, 0, 4);
	$strMonth = substr($strIsoDate, 4, 2);
	$strDay = substr($strIsoDate, 6, 2);
	// T
	$strHour = substr($strIsoDate, 9, 2);
	$strMinute = substr($strIsoDate, 11, 2);
	$strSecond = substr($strIsoDate, 13, 2);

	return mktime($strHour, $strMinute, $strSecond, $strMonth, $strDay, $strYear);
}

function FormatDate($strDate)
{
	$strYear = substr($strDate, 0, 4);
	$strMonth = substr($strDate, 4, 2);
	$strDay = substr($strDate, 6, 2);
	
	$strHour = substr($strDate, 9, 2);
	$strMinute = substr($strDate, 11, 2);

	return $strYear.'-'.$strMonth.'-'.$strDay.' '.$strHour . ':' . $strMinute;
}

// Crée des paragraphes xhtml à partir d'un texte comportant des lignes vides
function MakePs($str)
{
	$strTitle = '';
	$str = XhtmlSpecialChars($str);
	$astrPs = explode("\n\n", $str);

	$nPNum = 0;
	$strXhtml = '';
	foreach($astrPs as $strP)
	{
		$strP = trim($strP);
		$strXhtml .= '<p>' . $strP . '</p>' . chr(10);
		$nPNum++;
	}
	return $strXhtml;
}
////////////////////////////////////////////////////////////
function microtime_float()
{
   list($usec, $sec) = explode(' ', microtime());
   return ((float)$usec + (float)$sec);
}

// http://fr2.php.net/manual/fr/function.flock.php#46085
function AtomicFileAppend($strFileName, $data, $timeLimit = 300000, $staleAge = 5)
{
	ignore_user_abort(1);
	$lockDir = $strFileName . '.lock';

	if (is_dir($lockDir))
	{
		if ((time() - filemtime($lockDir)) > $staleAge)
		{
			rmdir($lockDir);
		}
	}

	$locked = @mkdir($lockDir);

	if ($locked === false)
	{
		$timeStart = microtime_float();
		do
		{
			if ((microtime_float() - $timeStart) > $timeLimit)
				break;
			$locked = @mkdir($lockDir);
		} while ($locked === false);
	}

	$bSuccess = false;

	if ($locked === true)
	{
		$hFile = @fopen($strFileName, 'ab');
		if( @fwrite($hFile, $data))
			$bSuccess = true;
		
		@fclose($hFile);
		rmdir($lockDir);
	}

	ignore_user_abort(0);
	return $bSuccess;
}

////////////////////////////////////////////////////////////
class CUnit
{
	var $m_strDate;
	var $m_strIp;
	var $m_strAuthor;
	var $m_strMessage;

	var $m_strError;

	function ReadNow($hFile, $nUnitNumber)
	{
		global $k_nUnitSize, $k_strUnitBegin, $k_strUnitEnd, $k_nAuthorMaxByteCount;

		$nOffset = $nUnitNumber * $k_nUnitSize;
		fseek($hFile, $nOffset);
				
		$nFilePos0 = ftell($hFile);
		
		$nMessageMaxByteCount = GetMessageMaxByteCount();
	
		$strUnitBegin = fread($hFile, strlen($k_strUnitBegin));
		$strDate = fread($hFile, 16);
		$strIp = fread($hFile, 16);
		$strAuthor = fread($hFile, $k_nAuthorMaxByteCount + 1);
		$strMessage = fread($hFile, $nMessageMaxByteCount + 1);
		$strUnitEnd = fread($hFile, strlen($k_strUnitEnd));

		$nFilePos1 = ftell($hFile);

		if( strcmp($strUnitBegin , $k_strUnitBegin) != 0)
		{
			$this->m_strError = 'Erreur : fichier DB endommagé : en-tête d\'unité non trouvée.';
			return false;
		}

		if( strcmp($strUnitEnd , $k_strUnitEnd) != 0)
		{
			$this->m_strError = 'Erreur : fichier DB endommagé : pied d\'unité non trouvé.';
			return false;
		}

		$nReadSoFar = ($nFilePos1 - $nFilePos0);
		if( $nReadSoFar != $k_nUnitSize)
		{
			$this->m_strError = 'Erreur : fichier DB endommagé : taille lue incohérente avec la taille d\'une unité. ' . "r=$nReadSoFar, s=$k_nUnitSize.";
			return false;
		}

		$this->m_strDate = $strDate;
		$this->m_strIp = $strIp;
		$this->m_strAuthor = $strAuthor;
		$this->m_strMessage = $strMessage;

		return true;
	}

	function EchoNow()
	{
		$strAuthor = trim($this->m_strAuthor);
		$strAuthor = XhtmlSpecialChars($strAuthor);

		echo '<div class="OneUnit">' . chr(10);
		echo '<div class="Message">' . chr(10);
		echo MakePs($this->m_strMessage);
		echo '</div>' . chr(10);
		echo '<div class="Author">' . $strAuthor . ' - ' . FormatDate($this->m_strDate) . '</div>' . chr(10);
		echo '</div>' . chr(10) . chr(10);
	}
}

function IsFlood($strNextDate, $strNextIp)
{
	global $k_strDbFileName, $k_nUnitSize;
	if( !file_exists($k_strDbFileName))
	{
		return array(false, '');
	}

	$strNextDate = trim($strNextDate); // Retire le padding
	$strNextIp = trim($strNextIp); // Retire le padding

	$nDbFileSize = filesize($k_strDbFileName);
	$nCount = $nDbFileSize / $k_nUnitSize;
	if( $nCount == 0)
		return array(false, '');

	$hFile = fopen($k_strDbFileName, 'rb');
	if( $hFile === false)
	{
		return array(false, 'IsFlood : Impossible d\'ouvrir le fichier DB');
	}
	
	$timeNext = IsoToTimestamp($strNextDate);

	$bFlood = false;
	for($iUnit = $nCount - 1; $iUnit >= 0; --$iUnit)
	{
		$unit = new CUnit;
		$unit->ReadNow($hFile, $iUnit);

		$timeUnit = IsoToTimestamp($unit->m_strDate);
		$timeDif = $timeNext - $timeUnit;
				
		if( $timeDif >= 60)
		{
			// On peut arrêter le parcours
			break;
		}

		$strUnitIp = trim($unit->m_strIp); // Retire \n
		if( strcmp( $strUnitIp, $strNextIp) == 0)
		{
			$bFlood = true;
			break;
		}
	}
	fclose($hFile);
	return array($bFlood, '');
}

if( isset($_GET['store']) && isset($_POST['UnitAuthor']) && isset($_POST['UnitMessage']) )
{
	$strGotAuthor = $_POST['UnitAuthor'];
	$strGotMessage = $_POST['UnitMessage'];

	//////////////////////////////////////////////////////////
	// On nettoie un peu ce qui vient de nous arriver
	$strGotAuthor = trim($strGotAuthor);
	$strGotMessage = trim($strGotMessage);

	// Remplace les fins de ligne exotiques par une fin de ligne normale
	$strGotMessage = str_replace("\r\n", "\n", $strGotMessage); // Façon DOS
	$strGotMessage = str_replace("\r", "\n", $strGotMessage); // Façon vieux Mac

	// L'option super chiante de PHP qui ajoute des \ aux apostrophes et guillemets
	if( get_magic_quotes_gpc())
	{
		// On retire les \ génants
		$strGotAuthor = stripslashes($strGotAuthor);
		$strGotMessage = stripslashes($strGotMessage);
	}
	//////////////////////////////////////////////////////////

	if( empty($strGotAuthor))
	{
		$strError = "Veuillez spécifier un nom.";
		$strUnitMessage = $strGotMessage;
	}
	else if( empty($strGotMessage))
	{
		$strError = "Veuillez spécifier un message.";
		$strUnitAuthor = $strGotAuthor;
	}
	else
	{
		$nMessageMaxByteCount = GetMessageMaxByteCount();

		$strGotAuthor = PadUtf8($strGotAuthor, $k_nAuthorMaxByteCount);
		$strGotMessage = PadUtf8($strGotMessage, $nMessageMaxByteCount);

		if( empty($strGotAuthor))
		{
			$strError = "Nom d\'autheur invalide (séquence UTF-8 incohérente).";
		}
		if( empty($strGotMessage))
		{
			$strError = "Message invalide (séquence UTF-8 incohérente).";
		}

		if( empty($strError))
		{
			// Pas d'erreur
			$timeNow = time();
			$strDate = date('Ymd', $timeNow) . 'T' . date('His', $timeNow);
			$strIp = $_SERVER['REMOTE_ADDR'];
			$strIp = str_pad($strIp, 15);

			list($bFlood, $strError) = IsFlood($strDate, $strIp);

			if( !$bFlood && empty($strError))
			{
				// Stockage
				$strData = $k_strUnitBegin;
				$strData .= $strDate . chr(10);
				$strData .= $strIp . chr(10);
				$strData .= $strGotAuthor . chr(10);
				$strData .= $strGotMessage . chr(10);
				$strData .= $k_strUnitEnd;

				if( AtomicFileAppend($k_strDbFileName, $strData) === false)
				{
					$strError = 'Impossible d\'écrire dans le fichier ' . $k_strDbFileName . '. ';
					$strError .= 'Conflit éventuel, patientez quelques secondes et revalidez votre message. ';
					$strError .= 'Si vous êtes l\'administrateur du site, vérifiez que les droits sont correctement fixés.';

					$strUnitAuthor = $strGotAuthor;
					$strUnitMessage = $strGotMessage;
				}
				else
				{
					// Ok
					clearstatcache();

					$strScriptName = $_SERVER['SCRIPT_NAME'];
					header("Location: $strScriptName");
					exit();
				}
			}
			else if( $bFlood)
			{
				$strError = 'Un message dont vous êtes l\'auteur (ou quelqu\'un de votre réseau local) a été envoyé il y a moins d\'une minute. ';
				$strError .= 'Patientez quelques instants et revalidez votre message.';

				$strUnitAuthor = $strGotAuthor;
				$strUnitMessage = $strGotMessage;
			}
		}
	}
}

$strUnitAuthor = XhtmlSpecialChars($strUnitAuthor);
$strUnitMessage = XhtmlSpecialChars($strUnitMessage);

//////////////////////////////////////////////////////////////////
// Ça c'est pour retransférer la page complètement si le navigant rafraichit la page
header('Expires: Mon, 10 Dec 1998 01:00:00 GMT');
header('Last-Modified: ' . gmdate('D, d M Y H:i:s') . ' GMT');
header('Cache-Control: no-cache, must-revalidate');

//////////////////////////////////////////////////////////////////
// Ligne à mettre au cas où on ne déssert pas en vrai xhtml
$strMetaCharset = '';

$bValidator = !(stristr($_SERVER['HTTP_USER_AGENT'], 'W3C_Validator') === false);

// Indique un type Mime correct si le client accepte le vrai xhtml
if( stristr($_SERVER['HTTP_ACCEPT'], 'application/xhtml+xml') || $bValidator) 
{
	header('Content-Type: application/xhtml+xml; charset=UTF-8');
	echo '<?xml version="1.0" encoding="UTF-8"?>' . "\n";
}
else 
{
	header('Content-Type: text/html; charset=UTF-8');
	$strMetaCharset = '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />' . "\n\t";
}
//////////////////////////////////////////////////////////////////
?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="fr" lang="fr">
<head>
	<?=$strMetaCharset?><title>Psy Guest Book</title>
	<link rel="stylesheet" type="text/css" href="psygb.css" title="psygb" />
</head>
<body>

<h1>Psy Guest Book</h1>

<?php
if( !file_exists($k_strDbFileName))
{
	echo '<p>Aucun message</p>'.chr(10);
}
else
{
	$nDbFileSize = filesize($k_strDbFileName);
	$nCount = $nDbFileSize / $k_nUnitSize;

	if( $nCount >= 2)
	{
		echo '<p class="MessageCount">'.$nCount.' messages</p>'.chr(10).chr(10);
	}
	
	$nLastPage = (int) ($nCount / $k_nMaxPerPage);
	$nLastPage += ($nCount % $k_nMaxPerPage) > 0 ? 1 : 0;

	// La page 1 est la page des plus anciens messages
	$nPage = 1;
	if( isset($_GET['page']))
	{
		$nPage = (int) $_GET['page'];
		if( $nPage < 1)
		{
			$nPage = 1;
		}
		else if( $nPage > $nLastPage)
		{
			$nPage = $nLastPage;
		}
	}
	else
	{
		// On va à la dernière page
		$nPage = $nLastPage;
	}

	$hFile = fopen($k_strDbFileName, 'rb');
	if( $hFile && $nCount > 0)
	{
		// Numéro du message tout en haut de la page (qu'il existe ou non)
		$nTopUnit = ((($nPage - 1) * $k_nMaxPerPage) + ($k_nMaxPerPage - 1));

		// Le message le plus récent se trouve à la fin du fichier
		$unit = new CUnit;
		for($iUnit = $nTopUnit; $iUnit > ($nTopUnit - $k_nMaxPerPage); --$iUnit)
		{
			if( $iUnit >= $nCount)
				continue;
			if( $unit->ReadNow($hFile, $iUnit) === false)
			{
				$strError = $unit->m_strError;
				break;
			}
			$unit->EchoNow();
		}
		fclose($hFile);
		
		////////////////////////////////////
		echo '<p class="Pages">Pages : ';
		$nPageBlock = (int) (($nPage - 1) / $k_nMaxPagePerBlock);
		
		//////////////////////////////////////////////
		// Écrit le lien vers le bloc de gauche
		$nLastPageBlock = (int) (($nLastPage - 1) / $k_nMaxPagePerBlock);
		if( $nPageBlock < $nLastPageBlock)
		{
			$nLink = (($nPageBlock + 1) * $k_nMaxPagePerBlock) + 1;
			$strLink = '<a class="PrevNext" href="?page=' . $nLink.'">'.($nLink).'</a> ';
			echo $strLink;
		}
		//////////////////////////////////////////////

		for($i = $k_nMaxPagePerBlock; $i >= 1; --$i)
		{
			$nLink = ($nPageBlock * $k_nMaxPagePerBlock) + $i;
			if( $nLink <= $nLastPage)
			{
				$strLink = '<a href="?page=' . $nLink.'">'.($nLink).'</a> ';
				if( $nLink == $nPage)
				{
					$strLink = '<strong>' . $nLink . '</strong> ';
				}
				echo $strLink;
			}
		}
		//////////////////////////////////////////////
		// Écrit le lien vers le bloc de droite
		if( $nPageBlock > 0)
		{
			$nLink = (($nPageBlock - 1) * $k_nMaxPagePerBlock) + $k_nMaxPagePerBlock;
			$strLink = '<a class="PrevNext" href="?page=' . $nLink.'">'.($nLink).'</a> ';
			echo $strLink;
		}
		//////////////////////////////////////////////
		echo '</p>'.chr(10);
	}
}

if( !empty($strError))
{
	echo '<p>' . $strError . '</p>' . chr(10).chr(10);
}
?>

<form id="UnitEdit" method="post" action="psygb.php?store=1">

<fieldset>
<p>
	<label for="UnitAuthor">Votre nom : </label>
	<input id="UnitAuthor" name="UnitAuthor" type="text" size="32" maxlength="26" value="<?=$strUnitAuthor?>" />
</p>

<p>
	<label for="UnitMessage">Votre message : </label>
	<textarea id="UnitMessage" name="UnitMessage" cols="44" rows="6"><?=$strUnitMessage?></textarea>
</p>

<p><input id="UnitSubmit" type="submit" value="Valider" /></p>

</fieldset>

</form>

<p id="GbFooter"><a href="http://psydk.org/psygb">psygb 1.1</a> - <a href="http://validator.w3.org/check/referer">xhtml 1.0</a></p>

</body>
</html>
